import logging
from pathlib import Path
from typing import Callable, List, Optional, TypeVar

import typer
from json_streams import jsonlib
from tabulate import tabulate

from karp import lex
from karp.foundation.commands import CommandBus
from karp.foundation.value_objects import unique_id
from karp.lex import commands as lex_commands
from karp.search import commands as search_commands
from karp.lex.application.queries import (
    GetPublishedResources,
    ListEntryRepos,
    GetResources,
)
from karp.errors import ResourceAlreadyPublished

from karp.cliapp.utility import cli_error_handler, cli_timer
from karp.cliapp.typer_injector import inject_from_ctx


logger = logging.getLogger("karp")


subapp = typer.Typer()


T = TypeVar("T")


def choose_from(choices: List[T], choice_fmt: Callable[[T], str]) -> T:
    for i, choice in enumerate(choices):
        typer.echo(f"{i}) {choice_fmt(choice)}")
    while True:
        number = typer.prompt(f"Choose from above with (0-{len(choices)-1}):")
        return choices[int(number)]


@subapp.command()
@cli_error_handler
@cli_timer
def create(
    ctx: typer.Context,
    config: Path,
    entry_repo_id: Optional[str] = typer.Option(None, help="id for entry-repo"),
):
    bus = inject_from_ctx(CommandBus, ctx)
    if config.is_file():
        data = jsonlib.load_from_file(config)
        if not entry_repo_id:
            query = inject_from_ctx(ListEntryRepos, ctx)
            entry_repos = list(query.query())
            entry_repo = choose_from(
                entry_repos, lambda x: f"{x.name} {x.repository_type}"
            )
            entry_repo_uuid = entry_repo.entity_id
        else:
            entry_repo_uuid = entry_repo_id
        cmd = lex_commands.CreateResource.from_dict(
            data,
            user="local admin",
            entry_repo_id=entry_repo_uuid,
        )
        bus.dispatch(cmd)
        typer.echo(f"Created resource '{cmd.resource_id}' ({cmd.entity_id})")

    elif config.is_dir():
        typer.Abort("not supported yetls")
        # new_resources = resources.create_new_resource_from_dir(config)

        # for resource_id in new_resources:
        #     typer.echo(f"Created resource {resource_id}")


@subapp.command()
@cli_error_handler
@cli_timer
def set_entry_repo(
    ctx: typer.Context,
    resource_id: str,
    entry_repo_id: str = typer.Argument(..., help="id for entry-repo"),
    user: Optional[str] = typer.Option(None),
):
    bus = inject_from_ctx(CommandBus, ctx)
    entry_repo_uuid = unique_id.parse(entry_repo_id)
    cmd = lex_commands.SetEntryRepoId(
        resource_id=resource_id,
        entry_repo_id=entry_repo_uuid,
        user=user or "local admin",
    )
    bus.dispatch(cmd)


@subapp.command()
@cli_error_handler
@cli_timer
def update(ctx: typer.Context, config: Path):
    bus = inject_from_ctx(CommandBus, ctx)
    # if config.is_file():
    #     with open(config) as fp:
    #         new_resource = resources.update_resource_from_file(fp)
    #     new_resources = [new_resource]
    # elif config.is_dir():
    #     new_resources = resources.update_resource_from_dir(config)
    # else:
    #     typer.echo("Must give either --config or --config-dir")
    #     raise typer.Exit(3)  # Usage error
    # for (resource_id, version) in new_resources:
    #     if version is None:
    #         typer.echo(f"Nothing to do for resource '{resource_id}'")
    #     else:
    #         typer.echo(f"Updated version {version} of resource '{resource_id}'")


@subapp.command()
@cli_error_handler
@cli_timer
def publish(ctx: typer.Context, resource_id: str):
    bus = inject_from_ctx(CommandBus, ctx)
    try:
        cmd = lex_commands.PublishResource(
            resource_id=resource_id,
            message=f"Publish '{resource_id}",
            user="local admin",
        )
        bus.dispatch(cmd)
    except ResourceAlreadyPublished:
        typer.echo("Resource already published.")
    else:
        typer.echo(f"Resource '{resource_id}' is published ")


@subapp.command()
@cli_error_handler
@cli_timer
def reindex(ctx: typer.Context, resource_id: str):
    bus = inject_from_ctx(CommandBus, ctx)
    cmd = search_commands.ReindexResource(resource_id=resource_id)
    bus.dispatch(cmd)

    typer.echo(f"Successfully reindexed all data in {resource_id}")


# @cli.command("pre_process")
# @click.option("--resource_id", required=True)
# @click.option("--version", required=True)
# @click.option("--filename", required=True)
# @cli_error_handler
# @cli_timer
# def pre_process_resource(resource_id, version, filename):
#     resource = resourcemgr.get_resource(resource_id, version=version)
#     with open(filename, "wb") as fp:
#         processed = indexmgr.pre_process_resource(resource)
#         pickle.dump(processed, fp)


# @cli.command("publish_preprocessed")
# @click.option("--resource_id", required=True)
# @click.option("--version", required=True)
# @click.option("--data", required=True)
# @cli_error_handler
# @cli_timer
# def index_processed(resource_id, version, data):
#     with open(data, "rb") as fp:
#         try:
#             loaded_data = pickle.load(fp)
#             resourcemgr.publish_resource(resource_id, version)
#             indexmgr.reindex(resource_id, version=version, search_entries=loaded_data)
#         except EOFError:
#             click.echo("Something wrong with file")


# @cli.command("reindex_preprocessed")
# @click.option("--resource_id", required=True)
# @click.option("--data", required=True)
# @cli_error_handler
# @cli_timer
# def reindex_preprocessed(resource_id, data):
#     with open(data, "rb") as fp:
#         try:
#             loaded_data = pickle.load(fp)
#             indexmgr.reindex(resource_id, search_entries=loaded_data)
#         except EOFError:
#             click.echo("Something wrong with file")


@subapp.command("list")
@cli_error_handler
@cli_timer
def list_resources(
    ctx: typer.Context,
    show_published: Optional[bool] = typer.Option(True, "--show-published/--show-all"),
):
    if show_published:
        query = inject_from_ctx(GetPublishedResources, ctx)
    else:
        query = inject_from_ctx(GetResources, ctx)
    typer.echo(
        tabulate(
            [
                [resource.resource_id, resource.version, resource.is_published]
                for resource in query.query()
            ],
            headers=["resource_id", "version", "published"],
        )
    )


@subapp.command()
@cli_error_handler
@cli_timer
def show(ctx: typer.Context, resource_id: str, version: Optional[int] = None):
    repo = inject_from_ctx(lex.ReadOnlyResourceRepository, ctx)
    resource = repo.get_by_resource_id(resource_id, version=version)
    #     if version:
    #         resource = database.get_resource_definition(resource_id, version)
    #     else:
    #         resource = database.get_active_or_latest_resource_definition(resource_id)
    if not resource:
        typer.echo(
            "Can't find resource '{resource_id}', version '{version}'".format(
                resource_id=resource_id,
                version=version if version else "latest",
            )
        )
        raise typer.Exit(3)

    typer.echo(tabulate(((key, value) for key, value in resource.dict().items())))


@subapp.command()
@cli_error_handler
@cli_timer
def set_permissions(
    ctx: typer.Context,
    resource_id: str,
    version: int,
    level: str,
):
    bus = inject_from_ctx(CommandBus, ctx)


#     # TODO use level
#     permissions = {"write": True, "read": True}
#     resourcemgr.set_permissions(resource_id, version, permissions)
#     click.echo("updated permissions")


@subapp.command()
@cli_error_handler
@cli_timer
def export():
    raise NotImplementedError()


@subapp.command()
@cli_error_handler
@cli_timer
def delete(
    ctx: typer.Context,
    resource_id: str,
    user: Optional[str] = typer.Option(None),
    message: Optional[str] = typer.Option(None),
):
    bus = inject_from_ctx(CommandBus, ctx)
    cmd = lex_commands.DeleteResource(
        resource_id=resource_id,
        user=user or "local admin",
        message=message or "resource deleted",
    )
    resource = bus.dispatch(cmd)
    typer.echo(f"Deleted resource '{resource_id}' ({resource})")


def init_app(app):
    app.add_typer(subapp, name="resource")



from starlette.config import Config
from uvicorn.workers import UvicornWorker


config = Config()


class ConfigurableWorker(UvicornWorker):
    """
    Define a UvicornWorker that can be configured by modifying its class attribute.

    All of the command line options for uvicorn are potential configuration options
    (see https://www.uvicorn.org/settings/ for the complete list).

    """

    #: dict: Set the equivalent of uvicorn command line options as keys.
    CONFIG_KWARGS = {
        "root_path": config("SCRIPT_NAME", default=""),
        "proxy_headers": True,
    }



import dataclasses
import typing
from typing import Dict, Iterable, List, Optional

import injector
from karp.foundation.events import EventBus

from karp.foundation.value_objects import UniqueId
from karp.foundation.commands import CommandBus
from karp.lex import EntryRepositoryUnitOfWorkFactory
from karp.lex.domain import entities as lex_entities
from karp.lex.application import repositories as lex_repositories
from karp.lex.application.queries import (
    ReadOnlyResourceRepository,
    ResourceDto,
)
from karp.tests.foundation.adapters import InMemoryUnitOfWork


@dataclasses.dataclass
class UnitTestContext:
    container: injector.Injector
    command_bus: CommandBus


class InMemoryResourceRepository(lex_repositories.ResourceRepository):
    def __init__(self):
        super().__init__()
        self.resources = {}

    def check_status(self):
        pass

    def _save(self, resource):
        self.resources[resource.id] = resource

    # def _update(self, resource):
    #     r = self._by_id(resource.id)
    #     self.resources.discard(r)
    #     self.resources.add(resource)

    def _by_id(self, id_, *, version=None):
        return self.resources.get(id_)

    def _by_resource_id(self, resource_id):
        return next(
            (res for res in self.resources.values() if res.resource_id == resource_id),
            None,
        )

    def __len__(self):
        return len(self.resources)

    def _get_published_resources(self) -> typing.Iterable[lex_entities.Resource]:
        return (res for res in self.resources.values() if res.is_published)

    def _get_all_resources(self) -> typing.Iterable[lex_entities.Resource]:
        return iter(self.resources.values())

    def resource_ids(self) -> typing.Iterable[str]:
        return (res.resource_id for res in self.resources)

    def num_entities(self) -> int:
        return sum(not res.discarded for res in self.resources.values())


class InMemoryReadResourceRepository(ReadOnlyResourceRepository):
    def __init__(self, resources: Dict):
        self.resources = resources

    def get_by_id(
        self, entity_id: UniqueId, version: Optional[int] = None
    ) -> Optional[ResourceDto]:
        resource = self.resources.get(entity_id)
        if resource:
            return self._row_to_dto(resource)
        return None

    def _get_by_resource_id(self, resource_id: str) -> Optional[ResourceDto]:
        return next(
            (
                self._row_to_dto(res)
                for res in self.resources.values()
                if res.resource_id == resource_id
            ),
            None,
        )

    def _row_to_dto(self, res) -> ResourceDto:
        return ResourceDto(
            entity_id=res.entity_id,
            resource_id=res.resource_id,
            last_modified=res.last_modified,
            last_modified_by=res.last_modified_by,
            version=res.version,
            config=res.config,
            is_published=res.is_published,
            entry_repository_id=res.entry_repository_id,
            name=res.name,
            message=res.message,
            discarded=res.discarded,
        )

    def get_published_resources(self) -> Iterable[ResourceDto]:
        return (
            self._row_to_dto(res) for res in self.resources.values() if res.is_published
        )


class InMemoryEntryRepository(lex_repositories.EntryRepository):
    def __init__(self):
        super().__init__()
        self.entries = set()

    def check_status(self):
        pass

    def _save(self, entry):
        self.entries.add(entry)

    def _update(self, entry):
        r = self._by_id(entry.id)
        self.entries.discard(r)
        self.entries.add(entry)

    def _by_id(
        self,
        id,
        *,
        version=None,
        after_date=None,
        before_date=None,
        oldest_first=False,
    ):
        return next((r for r in self.entries if r.id == id), None)

    # def _by_entry_id(
    #     self,
    #     entry_id,
    #     *,
    #     version=None,
    #     after_date=None,
    #     before_date=None,
    # ):
    #     return next((r for r in self.entries if r.entry_id == entry_id), None)

    def __len__(self):
        return len(self.entries)

    def _create_repository_settings(self, *args):
        pass

    @classmethod
    def from_dict(cls, _):
        return cls()

    def all_entries(self) -> typing.Iterable[lex_entities.Entry]:
        yield from self.entries

    def num_entities(self) -> int:
        return sum(not e.discarded for e in self.entries)

    def by_referenceable(
        self, filters: Optional[Dict] = None, **kwargs
    ) -> list[lex_entities.Entry]:
        return []


class InMemoryEntryUnitOfWork(InMemoryUnitOfWork, lex_repositories.EntryUnitOfWork):
    def __init__(
        self,
        entity_id,
        name: str,
        config: typing.Dict,
        connection_str: typing.Optional[str],
        message: str,
        user: str,
        event_bus: EventBus,
    ):
        InMemoryUnitOfWork.__init__(self)
        lex_repositories.EntryUnitOfWork.__init__(
            self,
            entity_id=entity_id,
            name=name,
            config=config,
            connection_str=connection_str,
            message=message,
            event_bus=event_bus,
        )
        self._entries = InMemoryEntryRepository()
        # self.id = entity_id
        # self.name = name
        # self.config = config

    @property
    def repo(self) -> lex_repositories.EntryRepository:
        return self._entries


class InMemoryEntryUnitOfWork2(InMemoryUnitOfWork, lex_repositories.EntryUnitOfWork):
    def __init__(self, entity_id, name: str, config: typing.Dict):
        InMemoryUnitOfWork.__init__(self)
        lex_repositories.EntryUnitOfWork.__init__(
            self, entity_id=entity_id, name=name, config=config
        )
        self._entries = InMemoryEntryRepository()
        # self.id = entity_id
        # self.name = name
        # self.config = config

    @property
    def repo(self) -> lex_repositories.EntryRepository:
        return self._entries


class InMemoryResourceUnitOfWork(
    InMemoryUnitOfWork, lex_repositories.ResourceUnitOfWork
):
    def __init__(self, event_bus: EventBus):
        InMemoryUnitOfWork.__init__(self)
        lex_repositories.ResourceUnitOfWork.__init__(self, event_bus=event_bus)
        self._resources = InMemoryResourceRepository()

    @property
    def repo(self) -> lex_repositories.ResourceRepository:
        return self._resources


# class InMemoryEntryUowFactory(lex_repositories.EntryUowFactory):
#     def create(
#         self,
#         resource_id: str,
#         resource_config: typing.Dict,
#         entry_repository_settings,
#     ) -> lex_repositories.EntryUnitOfWork:
#         entry_uow = InMemoryEntryUnitOfWork(entry_repository_settings)
#         if "entry_repository_type" in resource_config:
#             entry_uow.repo.type = resource_config["entry_repository_type"]
#         if entry_repository_settings:
#             entry_uow.repo_settings = entry_repository_settings
#         return entry_uow


class InMemoryEntryUnitOfWorkCreator:
    @injector.inject
    def __init__(self, event_bus: EventBus) -> None:
        self.event_bus = event_bus

    def __call__(
        self,
        entity_id: UniqueId,
        name: str,
        config: Dict,
        connection_str: Optional[str],
        user: str,
        message: str,
        timestamp: float,
    ) -> lex_repositories.EntryUnitOfWork:
        return InMemoryEntryUnitOfWork(
            entity_id=entity_id,
            name=name,
            config=config,
            connection_str=connection_str,
            message=message,
            user=user,
            event_bus=self.event_bus,
        )


def create_entry_uow2(
    entity_id: UniqueId,
    name: str,
    config: Dict,
) -> lex_repositories.EntryUnitOfWork:
    return InMemoryEntryUnitOfWork2(
        entity_id=entity_id,
        name=name,
        config=config,
    )


class InMemoryEntryUowRepository(lex_repositories.EntryUowRepository):
    def __init__(self) -> None:
        super().__init__()
        self._storage: dict[UniqueId, dict] = {}

    def _save(self, entry_repo):
        self._storage[entry_repo.id] = entry_repo

    def _by_id(self, id_, *, version: Optional[int] = None):
        return self._storage.get(id_)

    def __len__(self):
        return len(self._storage)

    def num_entities(self) -> int:
        return sum(not er.discarded for er in self._storage.values())


class InMemoryEntryUowRepositoryUnitOfWork(
    InMemoryUnitOfWork, lex_repositories.EntryUowRepositoryUnitOfWork
):
    def __init__(
        self,
        event_bus: EventBus,
        entry_uow_factory: EntryRepositoryUnitOfWorkFactory,
    ) -> None:
        InMemoryUnitOfWork.__init__(self)
        lex_repositories.EntryUowRepositoryUnitOfWork.__init__(
            self,
            event_bus=event_bus,
            entry_uow_factory=entry_uow_factory,
        )
        self._repo = InMemoryEntryUowRepository()

    @property
    def repo(self) -> lex_repositories.EntryUowRepository:
        return self._repo


class InMemoryLexInfrastructure(injector.Module):
    @injector.provider
    @injector.singleton
    def entry_uow_repo_uow(
        self,
        event_bus: EventBus,
        entry_uow_factory: EntryRepositoryUnitOfWorkFactory,
    ) -> lex_repositories.EntryUowRepositoryUnitOfWork:
        return InMemoryEntryUowRepositoryUnitOfWork(
            event_bus=event_bus,
            entry_uow_factory=entry_uow_factory,
        )

    # @injector.provider
    # def entry_uow_factory(self) -> lex_repositories.EntryRepositoryUnitOfWorkFactory:
    #     return InMemoryEntryRepositoryUnitOfWorkFactory()
    @injector.multiprovider
    def entry_uow_creator_map(
        self,
    ) -> Dict[str, lex_repositories.EntryUnitOfWorkCreator]:
        return {"fake": InMemoryEntryUnitOfWorkCreator}

    @injector.provider
    @injector.singleton
    def resource_uow(self, event_bus: EventBus) -> lex_repositories.ResourceUnitOfWork:
        return InMemoryResourceUnitOfWork(event_bus=event_bus)

    @injector.provider
    @injector.singleton
    def resource_repo(
        self,
        resource_uow: lex_repositories.ResourceUnitOfWork,
    ) -> ReadOnlyResourceRepository:
        return InMemoryReadResourceRepository(
            resources=resource_uow.repo.resources,
        )


# def bootstrap_test_app(
#     resource_uow: lex_repositories.ResourceUnitOfWork = None,
#     entry_uows: lex_repositories.EntriesUnitOfWork = None,
#     entry_uow_factory: lex_repositories.EntryUowFactory = None,
#     search_service_uow: SearchServiceUnitOfWork = None,
#     entry_repo_repo_uow: lex_repositories.EntryRepositoryRepositoryUnitOfWork = None,
# ):
#     return bootstrap_message_bus(
#         resource_uow=resource_uow or InMemoryResourceUnitOfWork(),
#         entry_repo_repo_uow=entry_repo_repo_uow or InMemoryEntryRepositoryRepositoryUnitOfWork(),
#         entry_uows=entry_uows or lex_repositories.EntriesUnitOfWork(),
#         entry_uow_factory=entry_uow_factory or InMemoryEntryUowFactory(),
#         search_service_uow=search_service_uow or InMemorySearchServiceUnitOfWork(),
#         raise_on_all_errors=True
#     )



from karp.foundation import messagebus
from karp.lex.application import handlers as lex_handlers, repositories as lex_repositories
from karp.lex.domain import commands as lex_commands, events as lex_events
from karp.search.application import handlers as search_handlers
from karp.search.application.unit_of_work import SearchServiceUnitOfWork
from karp.lex.application import repositories


def bootstrap_message_bus(
    *,
    resource_uow: repositories.ResourceUnitOfWork,
    entry_uows: repositories.EntriesUnitOfWork,
    entry_uow_factory,  #: repositories.EntryUowFactory,
    search_service_uow: SearchServiceUnitOfWork,
    entry_repo_repo_uow: lex_repositories.EntryRepositoryRepositoryUnitOfWork,
    raise_on_all_errors: bool = False
) -> messagebus.MessageBus:
    return messagebus.MessageBus(
        command_handlers={
            lex_commands.CreateResource: lex_handlers.CreateResourceHandler(
                resource_uow=resource_uow,
                entry_repo_repo_uow=entry_repo_repo_uow,
            ),
            lex_commands.PublishResource: lex_handlers.PublishResourceHandler(
                resource_uow=resource_uow,
            ),
            lex_commands.UpdateResource: lex_handlers.UpdateResourceHandler(
                resource_uow=resource_uow,
            ),
            # Entry command handlers
            lex_commands.AddEntry: lex_handlers.AddEntryHandler(
                resource_uow=resource_uow,
                entry_uows=entry_uows,
            ),
            lex_commands.DeleteEntry: lex_handlers.DeleteEntryHandler(
                resource_uow=resource_uow,
                entry_uows=entry_uows,
            ),
            lex_commands.UpdateEntry: lex_handlers.UpdateEntryHandler(
                resource_uow=resource_uow,
                entry_uows=entry_uows,
            )
        },
        event_handlers={
            lex_events.ResourceCreated: [
                search_handlers.CreateIndexHandler(
                    search_service_uow=search_service_uow
                ),
            ],
            lex_events.ResourcePublished: [],
            lex_events.ResourceUpdated: [],
            lex_events.EntryAdded: [],
            lex_events.EntryDeleted: [],
            lex_events.EntryUpdated: [],
        },
        raise_on_all_errors=raise_on_all_errors
    )



"""Pytest entry point."""
import pytest

from starlette.config import environ

environ["TESTING"] = "True"
environ["ELASTICSEARCH_HOST"] = "localhost:9202"
environ["CONSOLE_LOG_LEVEL"] = "DEBUG"

from karp.tests import common_data, utils  # nopep8


@pytest.fixture
def json_schema_config():
    return common_data.CONFIG_PLACES



from sqlalchemy import create_engine, orm, pool


class Database:
    def __init__(self, db_url: str) -> None:
        kwargs = {}
        if str(db_url).startswith("sqlite"):
            kwargs["poolclass"] = pool.SingletonThreadPool
            kwargs['connect_args'] = {
                'check_same_thread': False
            }
        self._engine = create_engine(db_url, echo=True, future=True, **kwargs)
        self.session_factory = orm.sessionmaker(
            autocommit=False,
            autoflush=False,
            bind=self._engine,
        )

    def disconnect(self):
        pass

#    @contextmanager
#    def session(self) -> Callable[..., AbstractContextManager[Session]]:
#        session: Session = self._session_factory()
#        try:
#            yield session
#        except Exception:
#            logger.exception('Session rollback because of exception')
#            session.rollback()
#            raise
#        finally:
#            session.close()
